<?php

namespace Drupal\config_overlay\Plugin\ConfigFilter;

use Drupal\config_filter\Plugin\ConfigFilterBase;
use Drupal\Core\Config\ExtensionInstallStorage;
use Drupal\Core\Config\FileStorageFactory;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides an overlay configuration filter.
 *
 * This will turn the synchronization storage into an overlay of the shipped
 * configuration of all enabled extensions.
 *
 * When writing configuration this filter will skip any configuration that is
 * identical to configuration provided by an installed extension. Thus, only
 * configuration that has been added or modified will be written to the
 * synchronization directory. When reading configuration the extensions'
 * configuration is amended to the configuration in the directory, so that the
 * missing files are transparent and configuration import, for example, works as
 * expected.
 *
 * Because in contrast to the active configuration shipped configuration
 * generally does not contain a UUID or a default configuration hash, the 'uuid'
 * and '_core' keys are ignored when comparing the active configuration with the
 * extension configuration to determine whether or not they are equal. When
 * reading the extension configuration as part of the overlay the respective
 * UUIDs and hashes of the active configuration are amended automatically so
 * that the configuration import does not detect any differences relative to the
 * active configuration.
 *
 * Note that due to this the configuration system will not be able to detect
 * configuration that has been deleted and recreated with the same name when
 * this filter is active. Such configuration will be detected as an update. To
 * avoid this and enable support for recreating configuration, UUIDs must be
 * added to the shipped configuration in custom modules or installation
 * profiles.
 *
 * Note that optional configuration is currently not supported, so that any
 * shipped optional configuration that the configuration system has installed
 * will be exported even if this filter is active. This will be fixed in the
 * future.
 *
 * @ConfigFilter(
 *   id = "config_overlay",
 *   label = @Translation("Module and installation profile configuration overlay"),
 * )
 */
class ConfigOverlayFilter extends ConfigFilterBase implements ContainerFactoryPluginInterface {

  /**
   * The active configuration storage;
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $activeStorage;

  /**
   * The extension configuration storage.
   *
   * @var \Drupal\Core\Config\ExtensionInstallStorage
   */
  protected $extensionStorage;

  /**
   * Configuration keys to ignore in the source configuration.
   *
   * In case any of these keys are present in the source configuration but not
   * in the extension configuration, the configuration will still be considered
   * equal.
   *
   * @var string[]
   *
   * @todo Make this configurable.
   */
  protected $keysToIgnore = ['_core', 'uuid'];

  /**
   * Constructs an extension configuration filter.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Config\StorageInterface $active_storage
   *   The active configuration storage.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, StorageInterface $active_storage) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->setActiveStorage($active_storage);
    // The extension storage depends on the source storage, so it cannot be set
    // at this point.
    /* @see \Drupal\config_overlay\Plugin\ConfigFilter\ExtensionConfigFilter::setSourceStorage() */
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('config.storage')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function setSourceStorage(StorageInterface $storage) {
    parent::setSourceStorage($storage);

    // The extension storage depends on the source storage to read the list of
    // installed extensions so initialize it as soon as the source storage is
    // available. When this filter is created for a non-default collection of
    // the source storage, a respective collection of the extension storage will
    // be created explicitly.
    /* @see \Drupal\config_overlay\Plugin\ConfigFilter\ExtensionConfigFilter::filterCreateCollection() */
    if ($storage->getCollectionName() === StorageInterface::DEFAULT_COLLECTION) {
      $name = 'core.extension';
      $profile = NULL;
      // Prior to the very first configuration export the 'core.extension'
      // configuration does not exist in the source storage. In this case we
      // fall back to passing the active storage to the extension storage,
      // otherwise this filter would not work at all.
      /* @see \Drupal\Core\Config\ExtensionInstallStorage::getAllFolders */
      /* @var \Drupal\Core\Config\StorageInterface $storage */
      foreach ([$storage, $this->activeStorage] as $possible_storage) {
        if ($possible_storage->exists($name)) {
          $storage = $possible_storage;

          $extension_config = $storage->read($name);
          if (isset($extension_config['profile'])) {
            $profile = $extension_config['profile'];
          }

          break;
        }
      }

      $this->setExtensionStorage(new ExtensionInstallStorage(
        $storage,
        ExtensionInstallStorage::CONFIG_INSTALL_DIRECTORY,
        $storage->getCollectionName(),
        TRUE,
        $profile
      ));
    }
  }

  /**
   * Sets the active configuration storage.
   * 
   * @param \Drupal\Core\Config\StorageInterface $active_storage
   */
  protected function setActiveStorage(StorageInterface $active_storage) {
    $this->activeStorage = $active_storage;
  }

  /**
   * Sets the extension configuration storage.
   *
   * @param \Drupal\Core\Config\StorageInterface $extension_storage
   */
  protected function setExtensionStorage(StorageInterface $extension_storage) {
    $this->extensionStorage = $extension_storage;
  }

  /**
   * {@inheritdoc}
   */
  public function filterExists($name, $exists) {
    if ($exists && !$this->getSourceStorage()->read($name)) {
      return FALSE;
    }

    return $exists || $this->extensionStorage->exists($name);
  }

  /**
   * {@inheritdoc}
   */
  public function filterWriteEmptyIsDelete($name) {
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function filterDelete($name, $delete) {
    if ($delete
        && !$this->getSourceStorage()->exists($name)
        && $this->extensionStorage->exists($name)) {

      $sync_storage = FileStorageFactory::getSync();
      touch($sync_storage->getFilePath($name));
      return FALSE;
    }

    return $delete;
  }

  /**
   * {@inheritdoc}
   */
  public function filterRead($name, $data) {
    // If the configuration does not exist in the sync directory, look it up
    // in the extensions' configuration.
    if (!$data) {
      $data = $this->extensionStorage->read($name);
      // Amend the exported configuration with the UUID and hash of the active
      // configuration.
      if ($data && ($active_data = $this->activeStorage->read($name))) {
        $data = $this->addIgnoredData($data, $active_data);
      }
    }
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function filterReadMultiple(array $names, array $data) {
    // Look up any missing configuration in the extensions' configuration.
    $remaining_names = array_diff($names, array_keys($data));
    $data += $this->extensionStorage->readMultiple($remaining_names);
    // Amend the exported configuration with the UUID and hash of the active
    // configuration.
    $active_data = $this->activeStorage->readMultiple($remaining_names);
    foreach ($remaining_names as $remaining_name) {
      if (isset($data[$remaining_name]) && isset($active_data[$remaining_name])) {
        $data[$remaining_name] = $this->addIgnoredData($data[$remaining_name], $active_data[$remaining_name]);
      }
    }
    // The data does not need to be sorted, as FilteredStorage::readMultiple()
    // does that already.
    /* @see \Drupal\config_filter\Config\FilteredStorage::readMultiple() */
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function filterWrite($name, array $data) {
    $data_to_compare = $data;

    $extension_storage = $this->extensionStorage->read($name);

    foreach ($this->keysToIgnore as $key_to_ignore) {
      if (!isset($extension_storage[$key_to_ignore])) {
        unset($data_to_compare[$key_to_ignore]);
      }
    }

    // Do not write anything if the active configuration matches configuration
    // provided by an extension.
    if ($data_to_compare === $extension_storage) {
      return;
    }
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function filterListAll($prefix, array $data) {
    $data = array_merge($data, $this->extensionStorage->listAll($prefix));
    $data = array_filter($data, [$this->getFilteredStorage(), 'exists']);
    // The data does not need to be sorted, as FilteredStorage::listAll()
    // does that already.
    /* @see \Drupal\config_filter\Config\FilteredStorage::listAll() */
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function filterDeleteAll($prefix, $delete) {
    if ($delete && $this->extensionStorage->listAll($prefix)) {
      return FALSE;
    }
    return $delete;
  }

  /**
   * {@inheritdoc}
   */
  public function filterCreateCollection($collection) {
    $clone = clone $this;
    $clone->setActiveStorage($this->activeStorage->createCollection($collection));
    $clone->setExtensionStorage($this->extensionStorage->createCollection($collection));
    return $clone;
  }

  /**
   * Adds ignored data from the active configuration to the exported data.
   *
   * @param array $data
   *   The exported configuration data that is being filtered.
   * @param array $active_data
   *   The active configuration data.
   *
   * @return array
   *   The amended configuration data.
   */
  public function addIgnoredData(array $data, array $active_data) {
    foreach ($this->keysToIgnore as $key_to_ignore) {
      if (!isset($data[$key_to_ignore]) && isset($active_data[$key_to_ignore])) {
        $data[$key_to_ignore] = $active_data[$key_to_ignore];
      }
    }

    // Make sure that the ignored data is positioned in the same place in the
    // data array as it is in the active configuration so that strict equality
    // between the exported and active configuration can be achieved. The
    // intersection makes sure that other keys that are available in the active
    // configuration but not in the exported configuration are not merged.
    $data = array_intersect_key(array_merge($active_data, $data), $data);
    return $data;
  }

}
