<?php

namespace Drupal\Tests\config_overlay\Functional;

use Drupal\Component\Utility\Crypt;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Config\StorageInterface;
use Drupal\Tests\BrowserTestBase;

/**
 * Provides a base test class for testing Configuration Overlay.
 */
class ConfigOverlayTestingTest extends BrowserTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['config_overlay'];

  /**
   * The configuration storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $configStorage;

  /**
   * The configuration synchronization storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $configSyncStorage;

  /**
   * The configuration synchronization directory.
   *
   * @var string
   */
  protected $configSyncDirectory;

  /**
   * The serializer.
   *
   * @var \Drupal\Component\Serialization\SerializationInterface
   */
  protected $serializer;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The configuration manager.
   *
   * @var \Drupal\Core\Config\ConfigManagerInterface
   */
  protected $configManager;

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  protected function setUp() {
    parent::setUp();

    $this->configStorage = $this->container->get('config.storage');
    $this->configSyncStorage = $this->container->get('config.storage.sync');
    $this->configSyncDirectory = config_get_config_directory(CONFIG_SYNC_DIRECTORY);
    $this->serializer = $this->container->get('serialization.yaml');
    $this->moduleHandler = $this->container->get('module_handler');
    $this->configManager = $this->container->get('config.manager');
    $this->entityTypeManager = $this->container->get('entity_type.manager');
  }

  /**
   * Tests configuration export after profile installation.
   *
   * @throws \Exception
   */
  public function testConfigExport() {
    $files = $this->exportConfig();
    foreach ($this->getExpectedConfig() as $name => $config) {
      $this->assertArrayHasKey($name, $files);

      $uri = $files[$name]->uri;
      $this->assertSame($config, $this->readConfigFile($uri), 'Configuration name: ' . $name);

      unset($files[$name]);
    }

    // This is functionally equivalent to Inspector::assertAllNotEmpty(), but
    // yields a more expressive error message in case of failure.
    foreach ($files as $name => $file) {
      $this->assertSame([], $this->readConfigFile($file->uri), 'Unexpected configuration: ' . $name);
    }
  }

  /**
   * Exports configuration for all collections and returns the exported files.
   *
   * @return \stdClass[]
   *   An associative array of objects with 'uri', 'filename', and 'name'
   *   properties corresponding to the matched files. The array is keyed by the
   *   'name' property of the respective files.
   *
   * @see _drush_config_export()
   */
  protected function exportConfig() {
    $this->syncAllConfiguration($this->configStorage, $this->configSyncStorage);
    foreach ($this->configStorage->getAllCollectionNames() as $collection_name) {
      $source = $this->configStorage->createCollection($collection_name);
      $target = $this->configSyncStorage->createCollection($collection_name);
      $this->syncAllConfiguration($source, $target);
    }

    $extension = $this->serializer->getFileExtension();
    return file_scan_directory($this->configSyncDirectory, "/.\.$extension$/", ['key' => 'name']);
  }

  /**
   * Exports configuration for a specific collection.
   *
   * @param \Drupal\Core\Config\StorageInterface $source
   *   The source storage for the given collection.
   * @param \Drupal\Core\Config\StorageInterface $target
   *   The target storage for the given collection.
   */
  protected function syncAllConfiguration(StorageInterface $source, StorageInterface $target) {
    foreach ($source->listAll() as $config_name) {
      $target->write($config_name, $source->read($config_name));
    }
  }

  /**
   * Gets an array of expected configuration that will be exported.
   *
   * Due to ExtensionConfigFilter this should only contain files that have
   * overridden configuration. For those it should contain the entire
   * configuration file.
   *
   * @return array[]
   *   An array of expected configuration where the keys are the configuration
   *   names and the values are arrays which contain the expected configuration.
   *
   * @see \Drupal\Tests\config_overlay\Functional\ConfigOverlayTestingTest::getOverriddenConfig()
   * @see \Drupal\Tests\config_overlay\Functional\ConfigOverlayTestingTest::getUri()
   */
  protected function getExpectedConfig() {
    $overridden_config = $this->getOverriddenConfig();

    $initial_config = [];
    // Read the initial config for all configuration that is overridden and
    // merge in the overridden values below.
    foreach (array_keys($overridden_config) as $config_name) {
      if ($uri = $this->getUri($config_name)) {
        $data = $this->readConfigFile($uri);
        $initial_config[$config_name] = $this->processConfigData($config_name, $data);
      }
    }

    return NestedArray::mergeDeepArray([$initial_config, $overridden_config], TRUE);
  }

  /**
   * Returns the overridden configuration for this test.
   *
   * @return array[]
   *   An array of overridden configuration where the keys are the configuration
   *   names and the values are arrays which contain the overridden portions of
   *   the configuration.
   */
  protected function getOverriddenConfig() {
    $site_config = $this->configStorage->read('system.site');

    return [
      'core.extension' => [
        'module' => [
          'config_filter' => 0,
          'config_overlay' => 0,
          'dynamic_page_cache' => 0,
          'page_cache' => 0,
          'system' => 0,
          'user' => 0,
          $this->profile => 1000,
        ],
        'theme' => [
          'stable' => 0,
          'classy' => 0,
        ],
        'profile' => $this->profile,
      ],
      /* @see install_base_system() */
      'system.file' => [
        'path' => [
          'temporary' => $this->tempFilesDirectory,
        ],
      ],
      /* @see \Drupal\Core\Test\FunctionalTestSetupTrait::initConfig() */
      'system.date' => [
        'timezone' => [
          'default' => 'Australia/Sydney',
        ],
      ],
      'system.logging' => [
        'error_level' => ERROR_REPORTING_DISPLAY_VERBOSE,
      ],
      'system.mail' => [
        'interface' => [
          'default' => 'test_mail_collector',
        ],
      ],
      'system.performance' => [
        'css' => [
          'preprocess' => FALSE,
        ],
        'js' => [
          'preprocess' => FALSE,
        ],
      ],
      'system.site' => [
        'uuid' => $site_config['uuid'],
        /* @see \Drupal\Core\Test\FunctionalTestSetupTrait::installParameters() */
        'name' => 'Drupal',
        'mail' => 'simpletest@example.com',
      ],
    ];
  }

  /**
   * Gets the file URI for the given configuration name.
   *
   * @param string $config_name
   *   The configuration name to return the file URI for.
   *
   * @return string|false
   *   The file URI for the given configuration, or FALSE if no configuration
   *   file was found.
   */
  protected function getUri($config_name) {
    $extension = $this->serializer->getFileExtension();

    $base_directory = NULL;
    // Reverse the list of directories so that the profile directory comes first
    // so that any profile-provided configuration will be used instead of
    // the respective module-provided configuration.
    $directories = array_reverse($this->moduleHandler->getModuleDirectories());
    $directories[] = $this->root . '/core';

    foreach ($directories as $directory) {
      $uri = $directory . "/config/install/$config_name.$extension";
      if (file_exists($uri)) {
        return $uri;
      }
    }

    return FALSE;
  }

  /**
   * Reads a single configuration file.
   *
   * @param string $uri
   *   The URI of the configuration file.
   *
   * @return array
   *   An array of configuration data contained in the file.
   */
  protected function readConfigFile($uri) {
    return $this->serializer->decode(file_get_contents($uri));
  }

  /**
   * Processes configuration data so that it matches the exported state.
   *
   * @param string $config_name
   *   The configuration name.
   * @param array $data
   *   The configuration data.
   *
   * @return array
   *   The processed configuration data.
   */
  protected function processConfigData($config_name, array $data) {
    // Add the default config hash.
    /* @see \Drupal\Core\Config\ConfigInstaller::createConfiguration() */
    $data['_core'] = ['default_config_hash' => Crypt::hashBase64(serialize($data))];
    if ($this->configManager->getEntityTypeIdByName($config_name)) {
      $data = $this->processConfigEntityData($config_name, $data);
    }
    return $data;
  }

  /**
   * Processes configuration entity data so that it matches the exported state.
   *
   * @param string $config_name
   *   The configuration name.
   * @param array $data
   *   The configuration entity data.
   *
   * @return array
   *   The processed configuration entity data.
   */
  protected function processConfigEntityData($config_name, array $data) {
    // Add the UUID from the active configuration.
    $top_data = ['uuid' => $this->configStorage->read($config_name)['uuid']];

    // Configuration entities are ordered in a particular way when
    // exported, so we need to recreate that here.
    $top_properties = [
      'langcode',
      'status',
      'dependencies',
      'third_party_settings',
      '_core',
    ];
    foreach ($top_properties as $property) {
      if (isset($data[$property])) {
        $top_data[$property] = $data[$property];
      }
    }

    return $top_data + $data;
  }

  /**
   * Reads an optional configuration file from a module.
   *
   * @param string $module_name
   *   The name of the module with the optional configuration.
   * @param string $config_name
   *   The configuration name.
   *
   * @return array
   *   An array of configuration data.
   */
  protected function readOptionalConfig($module_name, $config_name) {
    $module = $this->moduleHandler->getModule($module_name);
    $module_path = $this->root . '/' . $module->getPath();

    $data = $this->readConfigFile("$module_path/config/optional/$config_name.yml");
    return $this->processConfigData($config_name, $data);
  }

}
